(function() {
	'use strict';
	ObjC.import('AppKit');

	// ### MODULE=Polyfills.js
	if (!String.prototype.endsWith) {
		String.prototype.endsWith = function(searchString, position) {
			var subjectString = this.toString();
			if (typeof position !== 'number' || !isFinite(position) || Math.floor(position) !== position || position > subjectString.length) {
				position = subjectString.length;
			}
			position -= searchString.length;
			var lastIndex = subjectString.indexOf(searchString, position);
			return lastIndex !== -1 && lastIndex === position;
		};
	}

	if (!String.prototype.includes) {
		String.prototype.includes = function(search, start) {
			if (typeof start !== 'number') {
				start = 0;
			}

			if (start + search.length > this.length) {
				return false;
			} else {
				return this.indexOf(search, start) !== -1;
			}
		};
	}

	if (!String.prototype.startsWith) {
		String.prototype.startsWith = function(searchString, position) {
			position = position || 0;
			return this.substr(position, searchString.length) === searchString;
		};
	}
	// ### END MODULE

	function compare(a, b) {
		return a < b ? -1 : a > b ? 1 : 0;
	}

	var PathNameUtils = (function() {
		return {
			addTrailingSlash: function(path) {
				return path.endsWith("/") ? path : path + "/";
			},

			appendBeforeFileExtension: function(path, value) {
				var i = this.indexOfFileExtension(path);
				return i > 0 ?
					path.substring(0, i) + value + path.substring(i) :
					path + value;
			},

			combinePathComponents: function(path1, path2) {
				if (path1.endsWith("/"))
					path1 = path1.substring(0, path1.length - 1);
				if (path2.startsWith("/"))
					path2 = path2.substring(1);
				return path1 + "/" + path2;
			},

			expandLeadingPathTilde: function(path) {
				return path.startsWith("~") ?
					ObjC.unwrap($(path).stringByExpandingTildeInPath) :
					path;
			},

			getFileExtension: function(path) {
				var i = this.indexOfFileExtension(path);
				return (i <= 0) ? "" : path.substring(i);
			},

			getHomeFolder: function(addTrailingSlash) {
				var result = this.expandLeadingPathTilde("~");
				if (addTrailingSlash)
					result = this.addTrailingSlash(result);
				return result;
			},

			getLastPathComponent: function(path) {
				path = this.removeTrailingSlash(path);
				var i = path.lastIndexOf("/");
				return i < 0 ? path : path.substring(i + 1);
			},

			getStandardizedPath: function(path, keepTrailingSlash) {
				var hadTrailingSlash = path.endsWith("/");
				var result = $(path).stringByStandardizingPath.js;
				if (keepTrailingSlash && hadTrailingSlash && !result.endsWith("/"))
					result = result + "/";
				return result;
			},

			indexOfFileExtension: function(path) {
				var i = path.lastIndexOf(".");
				return (i <= 0 || path[i - 1] === "/") ? -1 : i;
			},

			removeLastPathComponent: function(path) {
				path = this.removeTrailingSlash(path);
				var last = this.getLastPathComponent(path);
				if (!last)
					return path;
				return path.substring(0, path.length - last.length);
			},

			removeTrailingSlash: function(path) {
				if (path.endsWith("/"))
					return path.substring(0, path.length - 1);
				return path;
			},

			replaceFileExtension: function(path, newExt) {
				if (!newExt) {
					newExt = "";
				} else if (!newExt.startsWith(".")) {
					newExt = "." + newExt;
				}
				var i = this.indexOfFileExtension(path);
				return (i <= 0) ? path + newExt : path.substring(0, i) + newExt;
			}
		};
	})();

	var FileUtils = (function() {
		return {
			chooseFile: function(prompt, type) {
				try {
					var options = {};
					if (prompt) options.withPrompt = prompt;
					if (type) options.ofType = type;
					return this.getCurrentApp().chooseFile(options).toString();
				} catch (e) {
					return "";
				}
			},

			copyFile: function(fromPath, toPath, throwIfFail) {
				var error;
				var result = ObjC.unwrap(
					$.NSFileManager.defaultManager
					.copyItemAtPathToPathError(
						$(fromPath).stringByStandardizingPath,
						$(toPath).stringByStandardizingPath,
						error
					)
				);
				if (error || (!result && throwIfFail))
					throw Error('Could not copy "' + fromPath + '" to "' + toPath + '"');
				return result;
			},

			createFolder: function(path, createIntermediateDirectories) {
				var error = $();
				if (!$.NSFileManager.defaultManager
					.createDirectoryAtPathWithIntermediateDirectoriesAttributesError(
						$(path).stringByStandardizingPath,
						createIntermediateDirectories,
						$(), error)) {
					throw Error("Could not create folder '" + path + "'");
				}
			},

			createFolderIfNeeded: function(path, createIntermediateDirectories) {
				if (!this.folderExists(path))
					this.createFolder(path, createIntermediateDirectories);
			},

			fileExists: function(path) {
				var result = this.getFileOrFolderExists(path);
				return result.exists && result.isFile;
			},

			folderExists: function(path) {
				var result = this.getFileOrFolderExists(path);
				return result.exists && !result.isFile;
			},

			getCurrentApp: function() {
				var app = Application.currentApplication();
				app.includeStandardAdditions = true;
				return app;
			},

			getDirectoryContents: function(strPath, includeSubfolders, includeDotFiles) {
				var method = includeSubfolders ?
					"subpathsOfDirectoryAtPathError" :
					"contentsOfDirectoryAtPathError";
				var error;
				var contents =
					ObjC.deepUnwrap(
						$.NSFileManager.defaultManager[method](
							$(strPath)
							.stringByStandardizingPath, error
						)
					);
				if (error)
					throw Error('Could not get contents of "' + strPath + '"');
				if (!includeDotFiles) {
					contents = contents.filter(function(item) {
						return !item.startsWith(".") && !item.includes("/.");
					});
				}
				return contents;
			},

			getFileOrFolderExists: function(path) {
				var isDirectory = Ref();
				var exists = $.NSFileManager.defaultManager
					.fileExistsAtPathIsDirectory(path, isDirectory);
				return {
					exists: exists,
					isFile: !isDirectory[0]
				};
			},

			readTextFile: function(strPath) {
				var error;
				var str = ObjC.unwrap(
					$.NSString.stringWithContentsOfFileEncodingError(
						$(strPath).stringByStandardizingPath,
						$.NSUTF8StringEncoding,
						error
					)
				);
				if (error)
					throw Error('Could not read file "' + strPath + '"');
				return str;
			},

			trashFile: function(path, throwIfFail) {
				return this.trashFileOrFolder(path, true, throwIfFail);
			},

			trashFileOrFolder: function(path, isFile, throwIfFail) {
				var fileOrFolderText = isFile ? "file" : "folder";
				if (!path)
					throw Error("Path is required in trashFileOrFolder");
				if (isFile === undefined)
					throw Error("isFile is required in trashFileOrFolder");

				if (!isFile && path.endsWith("/"))
					path = path.substring(0, path.length - 1);

				var exists = this.getFileOrFolderExists(path);
				if (!exists.exists)
					return false;
				if (exists.isFile !== isFile) {
					if (throwIfFail)
						throw Error("Path to be deleted '" + path + "' is not a " +
							fileOrFolderText);
					return false;
				}

				var fileURL = $.NSURL.fileURLWithPathIsDirectory($(path), !isFile);
				var error;
				var result = ObjC.unwrap(
					$.NSFileManager.defaultManager
					.trashItemAtURLResultingItemURLError(
						fileURL,
						null,
						error
					)
				);
				if (error || (!result && throwIfFail))
					throw Error("Could not trash " + fileOrFolderText + "'" + path + "'");
				return result;
			},

			trashFolder: function(path, throwIfFail) {
				return this.trashFileOrFolder(path, false, throwIfFail);
			},

			writeTextFile: function(strContent, strPath) {
				var error;
				var str = $.NSString.alloc.initWithUTF8String(strContent);
				str.writeToFileAtomicallyEncodingError(
					$(strPath).stringByStandardizingPath,
					true,
					$.NSUTF8StringEncoding,
					null);
				if (error)
					throw Error('Could not write file "' + strPath + '"');
			},

		};
	})();

	var ClipboardUtils = (function() {
		return {
			clearContents: function() {
				/*jshint -W030 */
				$.NSPasteboard.generalPasteboard.clearContents;
				/*jshint +W030 */
			},

			clipboardContainsType: function(type) {
				return this.getClipboardTypes().indexOf(type) >= 0;
			},

			getClipboardString: function(throwIfNotFound) {
				return this.getClipboardStringForType(
					this.getPlainTextClipboardType(), throwIfNotFound);
			},

			getClipboardStringForType: function(type, throwIfNotFound) {
				if (this.clipboardContainsType(type))
					return ObjC.unwrap(
						$.NSPasteboard.generalPasteboard
						.stringForType(type));
				if (throwIfNotFound)
					throw Error("No clipboard data for specified type");
				return "";
			},

			getClipboardTypes: function() {
				var items = $.NSPasteboard.generalPasteboard.pasteboardItems;
				if (items.count === 0)
					return [];

				return ObjC.deepUnwrap(items.js[0].types);
			},

			getPlainTextClipboardType: function() {
				return $.NSPasteboardTypeString.js;
			},

			setClipboardString: function(str) {
				this.setClipboardStringForType(str, this.getPlainTextClipboardType());
			},

			setClipboardStringForType: function(str, type) {
				var clipboard = $.NSPasteboard.generalPasteboard;
				/*jshint -W030 */
				clipboard.clearContents;
				/*jshint +W030 */
				clipboard.setStringForType($(str), $(type));
			}
		};
	})();

	var PlistUtils = (function() {
		function readTextFile(strPath) {
			var error;
			var str = ObjC.unwrap(
				$.NSString.stringWithContentsOfFileEncodingError(
					$(strPath).stringByStandardizingPath,
					$.NSUTF8StringEncoding,
					error
				)
			);
			if (error)
				throw Error('Could not read file "' + strPath + '"');
			return str;
		}

		return {
			convertPlistPartToString: function(plistPart) {
				var data = $.NSPropertyListSerialization.dataWithPropertyListFormatOptionsError(
					$(plistPart), $.NSPropertyListXMLFormat_v1_0, 0, null);
				var nsstring = $.NSString.alloc.initWithDataEncoding(data, $.NSUTF8StringEncoding);
				return $(nsstring).js;
			},

			convertStringToPlist: function(str) {
				return ObjC.deepUnwrap(
					$.NSPropertyListSerialization.propertyListWithDataOptionsFormatError(
						$(str).dataUsingEncoding($.NSUTF8StringEncoding), 0, 0, null));
			},

			readPlistArrayTextFile: function(strPath) {
				// var str = readTextFile(strPath);
				// return this.convertStringToPlist(str);
				var strFullPath = $(strPath).stringByStandardizingPath;
				return ObjC.deepUnwrap($.NSArray.arrayWithContentsOfFile(strFullPath));
			},

			readPlistBinaryFile: function(path) {
				var data = $.NSData.dataWithContentsOfFile(path);
				return ObjC.deepUnwrap(
					$.NSPropertyListSerialization.propertyListWithDataOptionsFormatError(
						data, $.NSPropertyListBinaryFormat_v1_0, 0, null));
			},

			readPlistDictionaryTextFile: function(strPath) {
				var strFullPath = $(strPath).stringByStandardizingPath;
				return ObjC.deepUnwrap($.NSDictionary.dictionaryWithContentsOfFile(strFullPath));
			},

			writePlistTextFile: function(plist, strPath) {
				var str = this.convertPlistPartToString(plist);
				$(str).writeToFileAtomically($(strPath).stringByStandardizingPath, true);
			}
		};
	})();

	var KMEngine = (function() {
		var _engineApp;
		var _currentApp = Application.currentApplication();
		_currentApp.includeStandardAdditions = true;
	
		return {
			calculate: function(str) {
				return this.getEngineApp().calculate(str);
			},

			convertStringToPlist: function(str) {
				return ObjC.deepUnwrap(
					$.NSPropertyListSerialization.propertyListWithDataOptionsFormatError(
						$(str).dataUsingEncoding($.NSUTF8StringEncoding), 0, 0, null));
			},

			deleteVariable: function(name) {
				this.setVariable(name, "%Delete%");
			},

			doScript: function(uuidOrUniqueNameOrScript, parameter) {
				if (parameter) {
					this.getEngineApp().doScript(uuidOrUniqueNameOrScript, { withParameter: parameter });
				} else {
					this.getEngineApp().doScript(uuidOrUniqueNameOrScript);
				}
			},

			executing: function() {
				return this.getEngineApp().executing();
			},

			getActionsClipboardType: function() {
				return "com.stairways.keyboardmaestro.actionarray";
			},

			getAllMacrosSourceFileName: function() {
				return this.getAppSupportFolderName() + "Keyboard Maestro Macros.plist";
			},

			getAllVariableNames: function() {
				return this.getEngineApp().variables.name();
			},

			getAppSupportFolderName: function() {
				return _currentApp.pathTo('application support', { from: 'user domain' }) +
					"/Keyboard Maestro/";
			},

			getEngineAppName: function() {
				return "Keyboard Maestro Engine";
			},

			getEngineApp: function() {
				if (!_engineApp)
					_engineApp = Application(this.getEngineAppName());
				return _engineApp;
			},

			getHotKeys: function(asString, getAll) {
				return this.getEngineApp().gethotkeys({ asstring: !!asString, getall: !!getAll });
			},

			getHotKeysAsPlist: function(getAll) {
				return this.convertStringToPlist(this.getHotKeys(true, getAll));
			},

			getMacrosClipboardType: function() {
				return "com.stairways.keyboardmaestro.macrosarray";
			},

			getMacros: function(binary) {
				return this.getEngineApp().getmacros({
					asstring: !binary
				});
			},

			getMacrosAsPlist: function(binary) {
				return this.convertStringToPlist(this.getMacros(false));
			},

			getNamedClipboardInfo: function() {
				var path = this.getNamedClipboardsSourceFileName();
				var plist = this.readPlistBinaryFile(path);

				var result = plist.map(function(item) {
					return { name: item.Name, UID: item.UID };
				});
				result = result.sort(function(a, b) {
					if (a.name < b.name) return -1;
					if (a.name > b.name) return 1;
					return 0;
				});
				return result;
			},

			getNamedClipboardsSourceFileName: function() {
				return this.getAppSupportFolderName() +
					"Keyboard Maestro Clipboards.plist";
			},

			getVariable: function(name, required) {
				var kme = this.getEngineApp();
				var result;
				if (name.match(/^Instance|^Local/)) {
					var inst = _currentApp.systemAttribute("KMINSTANCE");
					result = kme.getvariable(name, {instance: inst});
				} else {
					result = kme.getvariable(name);
				}
				if (required && !result)
					throw new Error("Variable '" + name + "' is empty");
				return result;
			},

			playSound: function(file, soundEffect, volume) {
				var options = {};
				if (soundEffect !== undefined)
					options.soundeffect = soundEffect;
				if (volume !== undefined)
					options.volume = volume;
				this.getEngineApp().playSound(file, options);
			},

			processTokens: function(str) {
				return this.getEngineApp().processTokens(str);
			},

			reload: function() {
				this.getEngineApp().reload();
			},

			setVariable: function(name, value) {
				var kme = this.getEngineApp();
				var options = {to: value};
				if (name.match(/^Instance|^Local/)) 
					options.instance = _currentApp.systemAttribute("KMINSTANCE");
				kme.setvariable(name, options);
			},
		};

	})();

	var KMEditor = (function() {
		var _editorAppName = "Keyboard Maestro";
		var _editorApp;

		return {
			deleteMacro: function(uuidOrUniqueName) {
				this.getEditorApp().deletemacro(uuidOrUniqueName);
			},

			deleteMacroGroup: function(uuidOrUniqueName) {
				this.getEditorApp().deletemacrogroup(uuidOrUniqueName);
			},

			editMacro: function(uuidOrUniqueName) {
				this.getEditorApp().editmacro(uuidOrUniqueName);
			},

			enableMacroOrGroup: function(uuidOrUniqueName, enable) {
				this.getEditorApp().setmacroenable(uuidOrUniqueName, { enable: enable });
			},

			getEditorApp: function() {
				return _editorApp ? _editorApp : _editorApp = Application(_editorAppName);
			},

			getEditorAppName: function() {
				return _editorAppName;
			},

			getSelectedMacrosOrGroups: function() {
				return this.getEditorApp().selectedmacros();
			},

			importMacros: function(pathOrPlist) {
				this.getEditorApp().importmacros(pathOrPlist);
			},

			reload: function() {
				this.getEditorApp().reload();
			},
		};
	})();

	var KMLogger = (function(logVariableName) {
		var _logVariableName = logVariableName;
		_clear();

		function _clear() {
			KMEngine.deleteVariable(_logVariableName);
		}

		return {
			clear: function() {
				_clear();
			},

			log: function(msg) {
				var currentLog = KMEngine.getVariable(_logVariableName);
				KMEngine.setVariable(_logVariableName, currentLog + msg + "\n");
			}
		};
	})("Instance_Log");

	// ### MODULE FavoriteActionsUsageDataAdapter.js
	var FavoriteActionsUsageDataAdapter = (function(fileUtils, kmEngine) {
		var _fileUtils = fileUtils;
		var _fileId = "FavoriteActionsUsage";
		var _usageFilePath = kmEngine.getVariable("Instance_PreferencesFolderPath", true) + "/FavoriteActionsUsage.json";

		return {
			loadDataFromPrefs: function() {
				var text = _fileUtils.readTextFile(_usageFilePath);
				if (text) {
					var data;
					try {
						data = JSON.parse(text);
					} catch (e) {
						data = null;
					}
					if (data && data.id && data.id === _fileId)
						return data.data;
				}
				return {};
			},

			saveDataToPrefs: function(data) {
				var prefsData = {
					id: _fileId,
					version: "1.0",
					data: data
				};
				_fileUtils.writeTextFile(JSON.stringify(prefsData), _usageFilePath);
			}
		};
	});
	// ### END MODULE

	// ### MODULE=FavoriteActionsAdapter.js
	var FavoriteActionsAdapter = (function(actions) {
		var _actions = actions;
		var _deletedActions = [];

		function _getActionByName(actionName, throwIfNotFound) {
			var name = actionName.toLowerCase();
			var result = _actions.find(function(action) {
				return action.name.toLowerCase() === name;
			});
			if (!result && throwIfNotFound)
				throw Error("Could not find Action named '" + actionName + "'");
			return result;
		}

		function _getActionByUUID(actionUUID, throwIfNotFound) {
			var index = _getActionIndexByUUID(actionUUID, throwIfNotFound);
			return index >= 0 ? _actions[index] : null;
		}

		function _getActionIndexByUUID(actionUUID, throwIfNotFound) {
			var index = _actions.findIndex(function(action) {
				return action.uuid === actionUUID;
			});
			if (index < 0 && throwIfNotFound)
				throw Error("Could not find Action with UUID = '" + actionUUID + "'");
			return index;
		}

		function _renumberOrderNumbers() {

			function _compare(a, b) {
				return a < b ? -1 : a > b ? 1 : 0;
			}

			function _compareActions(a, b) {
				if (a.orderNumber) {
					if (b.orderNumber)
						return _compare(a.orderNumber, b.orderNumber);
					return -1;
				}
				if (b.orderNumber)
					return 1;
				return _compare(a.name.toLowerCase(), b.name.toLowerCase());
			}

			// _renumberOrderNumbers
			var sortedActions = _actions.concat().sort(_compareActions);
			for (var i = 0; i < sortedActions.length; i++) {
				var action = sortedActions[i];
				if (action.orderNumber > 0) {
					action.orderNumber = i + 1;
				} else {
					break;
				}
			}
		} // _renumberOrderNumbers

		return {
			actionNameExists: function(actionName) {
				var result = _getActionByName(actionName);
				return result ? true : false;
			},

			deleteAction: function(action) {
				var index = _actions.indexOf(action);
				_actions.splice(index, 1);
				action.deleted = true;
				_deletedActions.push(action);
				_renumberOrderNumbers();
			},

			getActionByName: function(actionName, throwIfNotFound) {
				return _getActionByName(actionName, throwIfNotFound);
			},

			getActionByUUID: function(actionUUID, throwIfNotFound) {
				return _getActionByUUID(actionUUID, throwIfNotFound);
			},

			getActionIndexByUUID: function(actionUUID, throwIfNotFound) {
				return _getActionIndexByUUID(actionUUID, throwIfNotFound);
			},

			getActions: function() {
				return _actions;
			},

			getJsonString: function() {
				return JSON.stringify(_actions.concat(_deletedActions));
			},

			getNextOrderNumber: function() {
				var result = 1;
				_actions.forEach(function(action) {
					if (action.orderNumber >= result)
						result = action.orderNumber + 1;
				});
				return result;
			},

			pinAction: function(actionUUID) {
				var action = _getActionByUUID(actionUUID, true);
				action.orderNumber = this.getNextOrderNumber();
			},

			pinActionToggle: function(actionUUID) {
				var action = _getActionByUUID(actionUUID, true);
				if (action.orderNumber > 0) {
					this.unpinAction(actionUUID);
				} else {
					this.pinAction(actionUUID);
				}
			},

			unpinAction: function(actionUUID) {
				var action = _getActionByUUID(actionUUID, true);
				delete action.orderNumber;
				_renumberOrderNumbers();
			}
		};
	});
	// ### END MODULE

	// ### MODULE=FavoriteActionsDataAdapter.js
	var FavoriteActionsDataAdapter = (function(pathNameUtils, fileUtils, kmEngine, preferencesFolderPath, existingActionsAdapter) {
		var _pathNameUtils = pathNameUtils;
		var _fileUtils = fileUtils;
		var _kmEngine = kmEngine;
		var _preferencesFolderPath = preferencesFolderPath;
		var _dataFilePath = _pathNameUtils.combinePathComponents(_preferencesFolderPath, "FavoriteActions.json");
		var _imagesPath = _pathNameUtils.combinePathComponents(_preferencesFolderPath, "Images");
		var _fileId = "FavoriteActionsData";
		var _data = existingActionsAdapter ? existingActionsAdapter : _loadData();

		function _compare(a, b) {
			return a < b ? -1 : a > b ? 1 : 0;
		}

		function _loadData() {
			if (!_fileUtils.folderExists(_preferencesFolderPath))
				throw Error("Preferences Folder does not exist: '" + _preferencesFolderPath + "'");
			if (!_fileUtils.fileExists(_dataFilePath))
				throw Error("Preferences Folder appears to be corrupted.\n" +
					"Preferences Folder: '" + _preferencesFolderPath +
					"'\nMissing file '" + _dataFilePath + "'");

			var data;
			try {
				data = JSON.parse(_fileUtils.readTextFile(_dataFilePath));
			} catch (e) {
				throw Error("Invalid or corrupted Favorite Actions Preferences File '" + _dataFilePath + "'");
			}

			if (!data.id || data.id !== _fileId)
				throw Error("Invalid or corrupted Favorite Actions Preferences File '" + _dataFilePath + "'");
			return FavoriteActionsAdapter(data.actions);
		}

		return {
			addAction: function(actionUUID, actionName, actionKeywords, actionScript, actionType, actionImageFilePath) {
				var action = {
					uuid: actionUUID,
					name: actionName,
					keywords: actionKeywords,
					script: actionScript,
					type: actionType
				};
				_data.getActions().push(action);
				this.updateActionImage(action, actionImageFilePath);
			},

			getActionsInfo: function(includeIsDisabled) {
				var actions = _data.getActions();
				var result = [];
				var macroActions = null;
				for (var i = 0; i < actions.length; i++) {
					var action = actions[i];
					var actionInfo = {
						uuid: action.uuid,
						name: action.name,
						keywords: action.keywords,
						type: action.type,
						imageFileName: action.imageFileName,
						orderNumber: action.orderNumber,
						sortName: action.sortName
					};
					result.push(actionInfo);
					if (includeIsDisabled && action.type === "macro") {
						if (!macroActions)
							macroActions = {};
						macroActions[action.script] = actionInfo;
					}
				}

				if (macroActions) {
					var plist = _kmEngine.getMacrosAsPlist();
					plist.forEach(function(group) {
						if (!group.macros) return;

						group.macros.forEach(function(macro) {
							var actionInfo = macroActions[macro.uid];
							if (actionInfo) {
								actionInfo.isDisabled = !macro.active || !macro.enabled;
							}
						});
					});
				}

				return result;
			},

			getData: function() {
				return _data;
			},

			getExistingImageFilePathForUUID: function(actionUUID) {
				var fileNames = _fileUtils.getDirectoryContents(_imagesPath);
				var result;
				fileNames.forEach(function(fileName) {
					if (fileName.startsWith(actionUUID) && fileName.endsWith(".jpg")) {
						if (result)
							throw Error("Multiple image files exist for UUID '" + actionUUID + "'");
						result = _pathNameUtils.combinePathComponents(_imagesPath, fileName);
					}
				});
				return result;
			},

			getNewImageFilePathForUUID: function(actionUUID) {
				return _pathNameUtils.combinePathComponents(_imagesPath, actionUUID + " " + Date.now() + ".jpg");
			},

			getAlphaSortedActionNames: function() {
				return _data.getActions().map(function(action) {
					return action.name;
				}).sort(function(a, b) {
					return _compare(a.toLowerCase(), b.toLowerCase());
				});
			},

			save: function() {
				var fileData = {
					id: _fileId,
					version: "1.0",
					actions: _data.getActions()
				};
				_fileUtils.writeTextFile(JSON.stringify(fileData), _dataFilePath);
			},

			updateActionImage: function(action, actionImageFilePath) {
				var existingImageFilePath = this.getExistingImageFilePathForUUID(action.uuid);
				if (existingImageFilePath)
					_fileUtils.trashFile(existingImageFilePath);

				if (actionImageFilePath && _fileUtils.fileExists(actionImageFilePath)) {
					var destImageFilePath = this.getNewImageFilePathForUUID(action.uuid);
					_fileUtils.copyFile(actionImageFilePath, destImageFilePath);
					action.imageFileName = _pathNameUtils.getLastPathComponent(destImageFilePath);
				} else {
					action.imageFileName = null;
				}
			},

			updateExistingAction: function(actionName, actionKeywords, actionScript, actionType, actionImageFilePath, updateImageOnly) {
				var existingAction = _data.getActionByName(actionName, true);
				if (updateImageOnly !== "1") {
					if (actionKeywords)
						existingAction.keywords = actionKeywords;
					existingAction.script = actionScript;
					existingAction.type = actionType;
				}
				this.updateActionImage(existingAction, actionImageFilePath);
			},

			updateFromActionsInfo: function(updatedActionsInfo, usageInfo) {
				var actions = _data.getActions();
				updatedActionsInfo.forEach(function(updatedInfo) {
					var existingAction = _data.getActionByUUID(updatedInfo.uuid, true);

					if (updatedInfo.deleted) {
						this.updateActionImage(existingAction);
						_data.deleteAction(existingAction);
						if (usageInfo[existingAction.uuid])
							delete usageInfo[existingAction.uuid];
					} else {
						existingAction.name = updatedInfo.name;

						if (updatedInfo.sortName) {
							existingAction.sortName = updatedInfo.sortName;
						} else {
							delete existingAction.sortName;
						}

						if (updatedInfo.keywords) {
							existingAction.keywords = updatedInfo.keywords;
						} else {
							delete existingAction.keywords;
						}

						if (updatedInfo.orderNumber) {
							existingAction.orderNumber = updatedInfo.orderNumber;
						} else {
							delete existingAction.orderNumber;
						}
					}
				}, this);
			}
		};
	});
	// ### END MODULE

	// ### MODULE=OptionsDataAdapter.js
	var OptionsDataAdapter = (function(fileUtils, kmEngine) {
		var _fileUtils = fileUtils;
		var _fileId = "Options";

		var _usageFilePath = kmEngine.getVariable("Instance_PreferencesFolderPath", true);
		if (!_usageFilePath.endsWith("/"))
			_usageFilePath += "/";
		_usageFilePath += "Options.json";

		return {
			loadDataFromPrefs: function() {
				var text = _fileUtils.readTextFile(_usageFilePath);
				if (text) {
					var data;
					try {
						data = JSON.parse(text);
					} catch (e) {
						data = null;
					}
					if (data && data.id && data.id === _fileId)
						return data.data;
				}
				return {};
			},

			saveDataToPrefs: function(data) {
				var prefsData = {
					id: _fileId,
					version: "1.0",
					data: data
				};
				_fileUtils.writeTextFile(JSON.stringify(prefsData), _usageFilePath);
			},

			saveSortOptionsToPrefs: function(sortOptionsData) {
				var data = this.loadDataFromPrefs();
				data.sortOptions = sortOptionsData;
				saveDataToPrefs(data);
			}
		};
	});
	// ### END MODULE

	function applyAction(dataAdapter) {
		var usageData = JSON.parse(KMEngine.getVariable("Instance_UpdatedUsageData"));
		if (usageData) {
			var adapter = FavoriteActionsUsageDataAdapter(FileUtils, KMEngine);
			adapter.saveDataToPrefs(usageData);
		}

		var actionName = KMEngine.getVariable("Instance_ActionName", true);
		var action = dataAdapter.getData().getActionByName(actionName, true);

		var response;
		switch (action.type) {
			case "macro":
				KMEngine.doScript(action.script);
				break;
			case "action":
				ClipboardUtils.setClipboardStringForType(action.script, KMEngine.getActionsClipboardType());
				break;
			default:
				throw Error("Unknown Action Type '" + action.type + "'");
		}

		KMEngine.setVariable("Instance_ActionType", action.type);
		return "OK";
	}

	function getNewActionFromClipboard(dataAdapter) {
		var plistString = ClipboardUtils.getClipboardStringForType(KMEngine.getActionsClipboardType());
		if (!plistString) {
			plistString = ClipboardUtils.getClipboardString(KMEngine.getActionsClipboardType());
			if (!plistString.startsWith(`<?xml version="1.0" encoding="UTF-8"?>`))
				plistString = "";
		}
		if (!plistString)
			return "You must select one or more actions first.";

		var plist = PlistUtils.convertStringToPlist(plistString);
		if (!plist || plist.length === 0)
			return "You must select one or more actions first.";

		var actionName = plist[0].ActionName;
		if (!actionName)
			actionName = plist[0].MacroActionType;

		return returnNewAction(dataAdapter, actionName, plistString, "action");
	}

	function getNewActionFromSelectedMacro(dataAdapter) {
		var macros = KMEditor.getSelectedMacrosOrGroups();
		if (!macros)
			return "You must select one or more macros first.";
		var macroUUID = macros[0];
		var macroName = KMEngine.processTokens("%MacroNameForUUID%" + macroUUID + "%");
		if (macroName === "Not Found")
			return "You must select one or more macros first.";
		return returnNewAction(dataAdapter, macroName, macroUUID, "macro");
	}

	function getOptions(preferencesFolderPath) {
		var optionsDataAdapter = OptionsDataAdapter(FileUtils, KMEngine);
		var options = optionsDataAdapter.loadDataFromPrefs();
		KMEngine.setVariable("Instance_Options", JSON.stringify(options));
		KMEngine.setVariable("Instance_UpdatedOptions", "");
		return "OK";
	}

	function getSelectData(dataAdapter, preferencesFolderPath) {
		var usageDataAdapter = FavoriteActionsUsageDataAdapter(FileUtils, KMEngine);
		var selectData = {
			actionsInfo: dataAdapter.getActionsInfo(true),
			usageInfo: usageDataAdapter.loadDataFromPrefs(),
		};
		KMEngine.setVariable("Instance_SelectData", JSON.stringify(selectData));
		KMEngine.setVariable("Instance_UpdatedUsageData", "");

		return getOptions(preferencesFolderPath);
	}

	function goToMacro(dataAdapter) {
		var actionName = KMEngine.getVariable("Instance_ActionName", true);
		var action = dataAdapter.getData().getActionByName(actionName, true);
		if (action.type !== "macro")
			throw Error("Action is not a macro, in goToMacro: '" + actionName + "'");
		KMEditor.editMacro(action.script);
		return "OK";
	}

	function separateWords(str) {
		str = str.replace(/([a-z])([A-Z])/g, '$1 $2');
		str = str.replace(/([A-Z])([A-Z][a-z])/g, '$1 $2');
		return str;
	}

	function makeUniqueActionName(dataAdapter, str) {
		var name = separateWords(str);
		if (dataAdapter.getData().actionNameExists(name)) {
			for (var j = 1; j < 100; j++) {
				var newName = name + " " + j.toString();
				if (!dataAdapter.getData().actionNameExists(newName)) {
					name = newName;
					break;
				}
			}
		}
		return name;
	}

	function returnNewAction(dataAdapter, actionName, actionScript, actionType) {
		actionName = makeUniqueActionName(dataAdapter, actionName);

		var actionNamesPickList = dataAdapter.getAlphaSortedActionNames().join("|");

		KMEngine.setVariable("Instance_ActionName", actionName);
		KMEngine.setVariable("Instance_Action", actionScript);
		KMEngine.setVariable("Instance_ActionType", actionType);
		KMEngine.setVariable("Instance_ActionNamesPickList", actionNamesPickList);

		return "OK";
	}

	function saveNewAction(dataAdapter) {
		var actionUUID = KMEngine.getVariable("Instance_NewUUID", true);
		var actionName = KMEngine.getVariable("Instance_ActionName", true);
		var actionKeywords = KMEngine.getVariable("Instance_ActionKeywords");
		var actionScript = KMEngine.getVariable("Instance_Action", true);
		var actionType = KMEngine.getVariable("Instance_ActionType", true);
		var actionImageFilePath = KMEngine.getVariable("Instance_ActionImageFilePath");

		var response;
		if (!actionName) {
			response = "missing";
		} else if (dataAdapter.getData().actionNameExists(actionName)) {
			response = "duplicate";
		} else {
			dataAdapter.addAction(actionUUID, actionName, actionKeywords, actionScript, actionType, actionImageFilePath);
			dataAdapter.save();
			response = "OK";
		}

		KMEngine.setVariable("Instance_Response", response);
		return "OK";
	}

	function saveUpdatedOptions(preferencesFolderPath) {
		var updatedOptions = KMEngine.getVariable("Instance_UpdatedOptions");
		if (!updatedOptions)
			return "OK";
		var optionsAdapter = OptionsDataAdapter(FileUtils, KMEngine);
		optionsAdapter.saveDataToPrefs(JSON.parse(updatedOptions));
		return "OK";
	}

	function updateExistingAction(dataAdapter) {
		var actionName = KMEngine.getVariable("Instance_ActionName", true);
		var actionKeywords = KMEngine.getVariable("Instance_ActionKeywords");
		var actionScript = KMEngine.getVariable("Instance_Action", true);
		var actionType = KMEngine.getVariable("Instance_ActionType", true);
		var actionImageFilePath = KMEngine.getVariable("Instance_ActionImageFilePath");
		var updateImageOnly = KMEngine.getVariable("Instance_UpdateImageOnly", true);

		dataAdapter.updateExistingAction(actionName, actionKeywords, actionScript, actionType, actionImageFilePath,
			updateImageOnly);
		dataAdapter.save();
		return "OK";
	}

	function updateFromActionsInfo(dataAdapter) {
		var usageInfoAdapter = FavoriteActionsUsageDataAdapter(FileUtils, KMEngine);
		var usageInfo = usageInfoAdapter.loadDataFromPrefs();
		var actionsInfo = JSON.parse(KMEngine.getVariable("Instance_ActionsInfo", true));
		dataAdapter.updateFromActionsInfo(actionsInfo, usageInfo);
		dataAdapter.save();
		usageInfoAdapter.saveDataToPrefs(usageInfo);
		return "OK";
	}

	var _request = "not defined yet"

	function run() {
		var preferencesFolderPath = KMEngine.getVariable("Instance_PreferencesFolderPath", true);
		var dataAdapter = FavoriteActionsDataAdapter(PathNameUtils, FileUtils, KMEngine, preferencesFolderPath);
		_request = KMEngine.getVariable("Instance_Request", true);
		switch (_request) {
			case "applyAction":
				return applyAction(dataAdapter);
			case "getActionsInfo":
				KMEngine.setVariable("Instance_ActionsInfo", JSON.stringify(dataAdapter.getActionsInfo(true)));
				return "OK";
			case "getNewActionFromClipboard":
				return getNewActionFromClipboard(dataAdapter);
			case "getNewActionFromSelectedMacro":
				return getNewActionFromSelectedMacro(dataAdapter);
			case "getOptions":
				return getOptions(preferencesFolderPath);
			case "getSelectData":
				return getSelectData(dataAdapter, preferencesFolderPath);
			case "goToMacro":
				return goToMacro(dataAdapter);
			case "saveNewAction":
				return saveNewAction(dataAdapter);
			case "saveUpdatedOptions":
				return saveUpdatedOptions(preferencesFolderPath);
			case "updateExistingAction":
				return updateExistingAction(dataAdapter);
			case "updateFromActionsInfo":
				return updateFromActionsInfo(dataAdapter);
			default:
				throw Error("Unknown Request '" + _request + "'");
		}
	}

	try {
		return run();
	} catch (e) {
		var msgs = [`Error: ${e.message}`];
		msgs.push('\nDetails:\n--------');
		msgs.push([`Module: Favorite Actions Core`]);
		msgs.push([`Request: ${_request}`]);
		msgs.push([`Line: ${e.line}`]);
		return msgs.join("\n");
	}
})();